/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include "misc.h"
#include "midi.h"


#ifdef MAEMO
#define FIRST_NOTE 36
#define NUM_NOTES 48
#else
#define FIRST_NOTE 0
#define NUM_NOTES 128
#endif
#define KEY_WIDTH (win_width/((NUM_NOTES*7)/12))
#define KEY_HEIGHT ((win_height*7)/10)

static Display *disp;
static int screen_num;
static int win_width, win_height;
static Window root_win;
static GC gc;
static XColor col_white;
static XColor col_black;
static XColor col_active;
static XColor col_divide;
static Window win;
static Atom x_delete_atom;

static void keyboard_mouse(int x, int y, int down);
static void keyboard_open(void);
static void keyboard_close(void);
static void keyboard_draw(void);
static unsigned char keyboard_lookup(unsigned int keycode);

void
keyboard_toggle(void)
{
	if (keyboard_x_fd == -1) {
		keyboard_open();
	} else {
		keyboard_close();
	}
}

void
keyboard_poll(void)
{
	XEvent evt;
		/* key_up_note holds the note value of the previous key up
		 * event, so that we can merge up/down pairs that happen in
		 * the same call to keyboard_poll() into a no-op.  the note
		 * up event should be considered to be queued, and is only
		 * actually executed if it is not followed by a down event. */
	unsigned char key_up_note = 255;

	while (disp && XPending(disp)) {
		XNextEvent(disp, &evt);
		if (evt.xany.window != win) {
			return;
		}
		switch (evt.type) {
			case Expose:
				keyboard_draw();
				break;
			case KeyPress:
			case KeyRelease: {
				unsigned char note;
				unsigned char vol;
				vol = (evt.type == KeyPress) ? 127 : 0;
				note = keyboard_lookup(evt.xkey.keycode);
				if (note != 255) {
					if (vol && key_up_note == note) {
			/* repeat - gobble the up event because it was a lie
			 * followed by a down event which we also gobble. */
						key_up_note = 255;
					} else if (!vol) {
						if (key_up_note != 255) {
							cmd_note(key_up_note, 0);
						}
						key_up_note = note;
					} else {
						cmd_note(note, vol);
					}
				}
			}	break;
			case ConfigureNotify:
				win_width = evt.xconfigure.width;
				win_height = evt.xconfigure.height;
			case ButtonPress:
				keyboard_mouse(evt.xbutton.x, evt.xbutton.y, 1);
				break;
			case ButtonRelease:
				keyboard_mouse(evt.xbutton.x, evt.xbutton.y, 0);
				break;
			case MotionNotify:
				keyboard_mouse(evt.xmotion.x, evt.xmotion.y, !!evt.xmotion.state);
				break;
			case ClientMessage:
				if (evt.xclient.data.l[0] == x_delete_atom) {
					keyboard_close();
				}
				break;
		}
	}
	if (key_up_note != 255) {
		cmd_note(key_up_note, 0);
	}
	if (disp) XFlush(disp);
}

static unsigned char
classify_note(int x, int y)
{
	unsigned char n;
	int kw = KEY_WIDTH;
	int hkw = KEY_WIDTH/3;
	int hkh = KEY_HEIGHT/2;

		/* first handle the octave... */
	n = 12 * (x / (KEY_WIDTH * 7));
	x %= (KEY_WIDTH * 7);

		/* then check for black keys... any leftovers are white keys */
	if ((y < hkh) && (x >= hkw*2) && (x < hkw*4)) {
		n += 1;		/* C# */
	} else if ((y < hkh) && (x >= hkw*5) && (x < hkw*7)) {
		n += 3;		/* D# */
	} else if ((y < hkh) && (x >= hkw*11) && (x < hkw*13)) {
		n += 6;		/* F# */
	} else if ((y < hkh) && (x >= hkw*14) && (x < hkw*16)) {
		n += 8;		/* G# */
	} else if ((y < hkh) && (x >= hkw*17) && (x < hkw*19)) {
		n += 10;	/* A# */
	} else if (x < kw) {
		n += 0;		/* C */
	} else if (x < kw*2) {
		n += 2;		/* D */
	} else if (x < kw*3) {
		n += 4;		/* E */
	} else if (x < kw*4) {
		n += 5;		/* F */
	} else if (x < kw*5) {
		n += 7;		/* G */
	} else if (x < kw*6) {
		n += 9;		/* A */
	} else if (x < kw*7) {
		n += 11;	/* B */
	} else {
		printf("fucking fuck\n");
	}
	n += FIRST_NOTE;

	return n;
}

static void
piano_mouse(int x, int y, int down)
{
	static unsigned char old_note = 255;
	int new_note = (down ? classify_note(x,y) : 255);

	if (down) {
		if (new_note == old_note) {
			/* sustain */
		} else {
			if (old_note != 255) {
				/* glissando */
				cmd_note(old_note, 0);
				old_note = 255;
			}
			cmd_note(new_note, 127);
			old_note = new_note;
		}
	} else {
		if (old_note != 255) {
			cmd_note(old_note, 0);
			old_note = 255;
		}
	}
}

static void
keyboard_mouse(int x, int y, int down)
{
	if (y > KEY_HEIGHT) {
		if (!down) {
			cmd_sysex(x/(win_width/10));
		}
	} else {
		piano_mouse(x,y,down);
	}
}

static void
magic_wm_hints(void)
{
	Atom protos[2];
	x_delete_atom = XInternAtom(disp, "WM_DELETE_WINDOW", 0);
	protos[0] = x_delete_atom;
	protos[1] = XInternAtom(disp, "WM_TAKE_FOCUS", 0);
	XSetWMProtocols(disp, win, protos, 2);
}

static void
keyboard_open(void)
{
	Colormap cm;

	if (!(disp = XOpenDisplay(getenv("DISPLAY")))) {
		fprintf(stderr,"Failed to connect to X server.  Is DISPLAY right?\n");
		return;
	}

	screen_num = DefaultScreen(disp);
	root_win = RootWindow(disp, screen_num);

	gc = XCreateGC(disp, root_win, 0, 0);
	XSetFillStyle(disp, gc, FillSolid);
	cm = DefaultColormap(disp, screen_num);
	col_white.red   = 0xFFFF;
	col_white.green = 0xFFFF;
	col_white.blue  = 0xFFFF;
	XAllocColor(disp, cm, &col_white);
	col_black.red   = 0x0000;
	col_black.green = 0x0000;
	col_black.blue  = 0x0000;
	XAllocColor(disp, cm, &col_black);
	col_active.red   = 0x0000;
	col_active.green = 0xFFFF;
	col_active.blue  = 0xFFFF;
	XAllocColor(disp, cm, &col_active);
	col_divide.red   = 0xBBBB;
	col_divide.green = 0xBBBB;
	col_divide.blue  = 0xBBBB;
	XAllocColor(disp, cm, &col_divide);

	win = XCreateSimpleWindow(disp,root_win,0,0,
			900, 100, 0,
			BlackPixel(disp,screen_num),
			WhitePixel(disp,screen_num));
	XStoreName(disp, win, "ggb keyboard");
	XSelectInput(disp, win, ExposureMask | ButtonPressMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask | ButtonMotionMask | StructureNotifyMask);
	XMapWindow(disp, win);
	magic_wm_hints();

	XFlush(disp);

	keyboard_x_fd = ConnectionNumber(disp);
}

static void
keyboard_close(void)
{
	XCloseDisplay(disp);
	keyboard_x_fd = -1;
	disp = NULL;
}

static void
keyboard_draw(void)
{
	int i;

	if (win_width == 0) return;

	for (i = FIRST_NOTE; i < FIRST_NOTE+NUM_NOTES; i++) {
		keyboard_note(i, -1);
	}
	/* draw sysex buttons */
	XSetForeground(disp, gc, col_black.pixel);
	for (i = 1; i < 10; i++) {
		int x = i*win_width/10;
		XDrawLine(disp, win, gc, x, KEY_HEIGHT, x, win_height-1);
	}
	XFlush(disp);
}

/* down takes magic value of -1 when drawing the original untainted keyboard
 * to delay the XFlush() call */
void
keyboard_note(unsigned char note, int down)
{
	int x;
	enum { white_left, white_middle, white_right, black } kind = white_left;
	int kw = KEY_WIDTH;
	int hkw = KEY_WIDTH/3;
	int kh = KEY_HEIGHT;
	int hkh = KEY_HEIGHT/2;

	note -= FIRST_NOTE;

	if (keyboard_x_fd == -1) return;

		/* octave offset: */
	x = (note / 12) * (KEY_WIDTH * 7);
	note %= 12;

	switch (note) {
		case 0: x += 0;           kind = white_left;   break;
		case 1: x += 0;           kind = black;        break;
		case 2: x += kw;          kind = white_middle; break;
		case 3: x += kw;          kind = black;        break;
		case 4: x += kw*2;        kind = white_right;  break;
		case 5: x += kw*3;        kind = white_left;   break;
		case 6: x += kw*3;        kind = black;        break;
		case 7: x += kw*4;        kind = white_middle; break;
		case 8: x += kw*4;        kind = black;        break;
		case 9: x += kw*5;        kind = white_middle; break;
		case 10: x += kw*5;        kind = black;        break;
		case 11: x += kw*6;       kind = white_right;  break;
	}

	if (down == 1) {
		XSetForeground(disp, gc, col_active.pixel);
	} else if (kind == black) {
		XSetForeground(disp, gc, col_black.pixel);
	} else {
		XSetForeground(disp, gc, col_white.pixel);
	}
	switch (kind) {
		case black:
			XFillRectangle(disp, win, gc, x+hkw*2, 0, hkw*2, hkh);
			break;
		case white_left:
			XFillRectangle(disp, win, gc, x, 0, hkw*2, hkh);
			XFillRectangle(disp, win, gc, x, hkh, kw, hkh);
			XSetForeground(disp, gc, col_divide.pixel);
			XDrawLine(disp, win, gc, x, 0, x, kh);
			break;
		case white_middle:
			XFillRectangle(disp, win, gc, x+hkw, 0, hkw, hkh);
			XFillRectangle(disp, win, gc, x, hkh, kw, hkh);
			XSetForeground(disp, gc, col_divide.pixel);
			XDrawLine(disp, win, gc, x, hkh, x, kh);
			break;
		case white_right:
			XFillRectangle(disp, win, gc, x+hkw, 0, hkw*2, hkh);
			XFillRectangle(disp, win, gc, x, hkh, kw, hkh);
			XSetForeground(disp, gc, col_divide.pixel);
			XDrawLine(disp, win, gc, x, hkh, x, kh);
			break;
	}

	if (down != -1)
		XFlush(disp);
}

#define KEYCODE_OFFSET 36;	/* XXX: this should really be user-definable as any multiple of 12 */

static const int keycodes[] = {
#if defined(KEYBOARD_LINUX)
/*	               lsh   z   s   x   d   c   f   v	*/
	-1, -1, -1, -1, 50, 52, 39, 53, 40, 54, 41, 55,
/*	 b   h   n   j   m   ,   l   .   ;   /   ' rsh	*/
	56, 43, 57, 44, 58, 59, 46, 60, 47, 61, 48, 62,
/*	tab  1   q   2   w   e   4   r   5   t   6   y	*/
	23, 10, 24, 11, 25, 26, 13, 27, 14, 28, 15, 29,
/*	 u   8   i   9   o   p   -   [   =   ] bsp   \	*/
	30, 17, 31, 18, 32, 33, 20, 34, 21, 35, 22, 51,
/*	esc F1  F2  F3  F4  F5  F6  F7  F8  F9 F10 F11	*/
	 9, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 95,
/*	F12						*/
	96
#elif defined(KEYBOARD_IBOOK)
/*	               lsh   z   s   x   d   c   f   v	*/
	-1, -1, -1, -1, 64, 14,  9, 15, 10, 16, 11, 17,
/*	 b   h   n   j   m   ,   l   .   ;   /   ' rsh	*/
	19, 12, 53, 46, 54, 51, 45, 55, 49, 52, 47, 44,
/*	tab  1   q   2   w   e   4   r   5   t   6   y	*/
	56, 26, 20, 27, 21, 22, 29, 23, 31, 25, 30, 24,
/*	 u   8   i   9   o   p   -   [   =   ] bsp   \	*/
	40, 36, 42, 33, 39, 43, 35, 41, 32, 38, 59, 50,
/*	esc F1  F2  F3  F4  F5  F6  F7  F8  F9 F10 F11	*/
	61,130,128,107,126,104,105,106,108,109,117,111,
/*	F12						*/
	119
#else
#error undefined keyboard
#endif
};

static unsigned char
keyboard_lookup(unsigned int keycode)
{
	int i;

	for (i = 0; i < sizeof keycodes / sizeof keycodes[0]; i++) {
		if (keycode == keycodes[i]) {
			return i + KEYCODE_OFFSET;
		}
	}
#if 0
	printf("unknown keycode: %d\n",keycode);
#endif
	return 255;
}
