/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: snd.c,v 1.45 2010-02-08 03:17:28 greg Exp $ */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "lang.h"
#include "snd.h"
#include "misc.h"
#include "tracker.h"
#include "midi.h"

static signed short *snd_compute(void);
static int *snd_fbuf = NULL;		/* stereo accum output buffer */

#if defined(ALSA_SOUND)

#include <alsa/asoundlib.h>

static snd_pcm_t *alsa_handle;
static int FRAG_SAMPLES = 1024;

#define STRINGIFY(x) STRINGIFY1(x)
#define STRINGIFY1(x) #x
#define SND_PCM(x) { int err; if ((err=snd_pcm_##x)<0) { \
		fprintf(stderr, "snd_pcm_" STRINGIFY(x) ": %s\n", snd_strerror(err)); \
		snd_off(); \
		return -1; \
	} }
int
snd_on(void)
{
	snd_pcm_hw_params_t *hwp;
	snd_pcm_sw_params_t *swp;
	unsigned rate = SEC_MULT;
	unsigned buffer_time = 150000; /* buffer length in us */
	unsigned period_time = 50000;
	unsigned buf_len;
	snd_pcm_t *h;
	int nfds;
	static struct pollfd pfd;
	snd_pcm_uframes_t size;

	snd_pcm_hw_params_alloca(&hwp);
	snd_pcm_sw_params_alloca(&swp);

	if (alsa_handle) {
		/* already inited */
		return 0;
	}

	SND_PCM(open(&alsa_handle, ALSA_FILE, SND_PCM_STREAM_PLAYBACK, 0))
	h = alsa_handle;
	SND_PCM(hw_params_any(h, hwp))
	SND_PCM(hw_params_set_access(h, hwp, SND_PCM_ACCESS_RW_INTERLEAVED))
	SND_PCM(hw_params_set_format(h, hwp, SND_PCM_FORMAT_S16))
	SND_PCM(hw_params_set_channels(h, hwp, 2))
	SND_PCM(hw_params_set_rate_near(h, hwp, &rate, 0))
	SND_PCM(hw_params_set_buffer_time_near(h, hwp, &buffer_time, 0))
	SND_PCM(hw_params_set_period_time_near(h, hwp, &period_time, 0))
	SND_PCM(hw_params(h, hwp))
#if 0
printf("buf_len=%d\n",buf_len);
{snd_output_t *o;
snd_output_stdio_attach(&o, stdout, 0);
snd_pcm_dump(h, o);
snd_pcm_dump_setup(h, o);
snd_pcm_hw_params_dump(hwp, o);
}
#endif /* 0 */
	SND_PCM(hw_params_get_rate(hwp, &rate, 0))
	/* XXX - if rate != 44100, maybe we have trouble */
	SND_PCM(hw_params_get_buffer_size(hwp, &size))
	buf_len = size;
	SND_PCM(hw_params_get_period_size(hwp, &size, 0))
	FRAG_SAMPLES = size;
	SND_PCM(sw_params_current(h, swp))
	SND_PCM(sw_params_set_start_threshold(h, swp, buf_len))
	SND_PCM(sw_params_set_avail_min(h, swp, FRAG_SAMPLES))
	SND_PCM(sw_params(h, swp))
	
	nfds = snd_pcm_poll_descriptors_count(h);
	if (nfds != 1) {
		fprintf(stderr,"alsa pcm wants more than one fd, fuck it\n");
		return -1;
	}
	SND_PCM(poll_descriptors(h, &pfd, 1))
	snd_fd = pfd.fd;

	return 0;
}

void
snd_out(void)
{
	int i;
	signed short *sndbuf;
	int pos = 0;

	sndbuf = snd_compute();
	while (alsa_handle && (pos < FRAG_SAMPLES)) {
		i = snd_pcm_writei(alsa_handle, sndbuf+pos, FRAG_SAMPLES-pos);
		if (i == -EAGAIN) {
			continue;
		}
		if (i <= 0) {
			snd_pcm_resume(alsa_handle);
			snd_pcm_prepare(alsa_handle);
			break;
		}
		pos += i;
	}
	if (snd_file_fd != -1) {
		write(snd_file_fd, sndbuf, FRAG_SAMPLES*4);
	}
}

void
snd_off(void)
{
	if (!alsa_handle) {
		return;
	}

	snd_pcm_close(alsa_handle);
	snd_fd = -1;
	alsa_handle = NULL;
}

void
snd_changedev(char *dev)
{
	fprintf(stderr, "Currently only supports default plug in ALSA.\n");
}


#elif defined(OSS_SOUND)



#include <sys/soundcard.h>

static int mic_fd = -1;
static char *curr_dev = NULL;

int
snd_on(void)
{
	int i;

	if (!curr_dev) curr_dev = SND_FILE;

	snd_fd = open(curr_dev, O_WRONLY);
	if (snd_fd < 0) {
		perror(curr_dev);
		return -1;
	}
	i = 0x00040000 | FRAG_LEN;
	ioctl(snd_fd,SNDCTL_DSP_SETFRAGMENT,&i);
	if (i != (0x00040000 | FRAG_LEN)) {
		perror("ioctl(SNDCTL_DSP_SETFRAGMENT,0x7ffe0000");
		return -1;
	}
	i = AFMT_S16_LE;                /* PORTABILITY -> _BE for big endian */
	ioctl(snd_fd,SNDCTL_DSP_SETFMT,&i);
	if (i != AFMT_S16_LE) {
		perror("ioctl(SNDCTL_DSP_SETFMT,AFMT_S16_LE");
		return -1;
	}
	i = 1;
	ioctl(snd_fd,SNDCTL_DSP_STEREO,&i);
	if (i != 1) {
		perror("ioctl(SNDCTL_DSP_STEREO,1");
		return -1;
	}
	i = 44100;
	ioctl(snd_fd,SNDCTL_DSP_SPEED,&i);
	if (i < 41895 || i > 46305) {   /* 5% tolerances */
		perror("ioctl(SNDCTL_DSP_SPEED,44100");
		return -1;
	}

	return 0;
}


void snd_off(void)
{
	close(snd_fd);
	snd_fd = -1;

	if (mic_fd != -1) {
		close(mic_fd);
		mic_fd = -1;
	}
}



/* adds the microphone input to sndbuf.. */
static void
snd_read_mic(signed short *sndbuf)
{
	signed short micbuf[FRAG_SAMPLES];
	int i;

	if (mic_fd == -1) {
		mic_fd = open(MIC_FILE, O_RDONLY);
		if (mic_fd < 0) {
			static int yelled = 0;
			if (!yelled) {
				perror(MIC_FILE);
				yelled = 1;
			}
			return;
		}

		/* halve the frag len because mic is not stereo */
		i = 0x00040000 | (FRAG_LEN >> 1);
		ioctl(mic_fd,SNDCTL_DSP_SETFRAGMENT,&i);
		if (i != (0x00040000 | (FRAG_LEN >> 1))) {
			perror("ioctl(SNDCTL_DSP_SETFRAGMENT,...)");
		}
		i = AFMT_S16_LE;                /* PORTABILITY -> _BE for big endian */
		ioctl(mic_fd,SNDCTL_DSP_SETFMT,&i);
		if (i != AFMT_S16_LE) {
			perror("ioctl(SNDCTL_DSP_SETFMT,AFMT_S16_LE)");
		}
		i = 0;
		ioctl(mic_fd,SNDCTL_DSP_STEREO,&i);
		if (i != 0) {
			perror("ioctl(SNDCTL_DSP_STEREO,0)");
		}
		i = 44100;
		ioctl(mic_fd,SNDCTL_DSP_SPEED,&i);
		if (i < 41895 || i > 46305) {   /* 5% tolerances */
			perror("ioctl(SNDCTL_DSP_SPEED,44100");
		}
	}
	read(mic_fd, micbuf, FRAG_SAMPLES*2);
	for (i = 0; i < FRAG_SAMPLES; i++) {
		signed short v = ((((signed int)micbuf[i]) * mic_level) / 32768);
		sndbuf[i*2+0] += v;
		sndbuf[i*2+1] += v;
	}
}

void
snd_out(void)
{
	signed short *sndbuf;

	if (!snd_enabled) {
		return;
	}

	sndbuf = snd_compute();

	if (mic_level) {
		snd_read_mic(sndbuf);
	} else if (mic_fd != -1) {
		close(mic_fd);
		mic_fd = -1;
	}

	if (snd_fd != -1) {
		write(snd_fd, sndbuf, FRAG_SAMPLES*4);
	}
	if (snd_file_fd != -1) {
		write(snd_file_fd, sndbuf, FRAG_SAMPLES*4);
	}
}

void
snd_changedev(char *dev)
{
	snd_cleanup();

	if (!dev || !*dev) {
		curr_dev = SND_FILE;	/* default device */
		printf("switching to default: %s\n", curr_dev);
	} else {
		curr_dev = safe_strdup(dev);
	}

	if (snd_init()) {
		printf("failed...reverting to default: %s\n", SND_FILE);
		curr_dev = SND_FILE;
		snd_init();
	}
}


#elif defined(COREAUDIO_SOUND)

#include <CoreServices/CoreServices.h>
#include <AudioUnit/AudioUnit.h>

static AudioUnit snd_out_unit;


static OSStatus snd_render(void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData);

int
snd_on(void)
{
	OSStatus err = noErr;
	ComponentDescription desc;
	Component comp;
	AURenderCallbackStruct input;
	AudioStreamBasicDescription sfmt;

	snd_fd = -1;
	sounds = NULL;

	/* open the default output unit */
	desc.componentType = kAudioUnitType_Output;
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;

	comp = FindNextComponent(NULL, &desc);
	if (comp == NULL) {
		fprintf(stderr, "FindNextComponent\n");
		return -1;
	}

	err = OpenAComponent(comp, &snd_out_unit);
	if (err) {
		fprintf(stderr, "OpenAComponent=%ld\n", err);
		return -1;
	}

	/* set up a callback function to generate output to the output unit */
	input.inputProc = snd_render;
	input.inputProcRefCon = NULL;
	err = AudioUnitSetProperty (snd_out_unit,
		kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input,
		0, &input, sizeof(input));
	if (err) {
		fprintf(stderr, "AudioUnitSetProperty-CB=%ld\n", err);
		return -1;
	}

	/* set up the format for the output stream */
	sfmt.mSampleRate = 44100;
	sfmt.mFormatID = kAudioFormatLinearPCM;
	sfmt.mFormatFlags = kAudioFormatFlagIsBigEndian | kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked;
	/* XXX - I'm not sure this is really correct.  I want ideally
	 * to send FRAG_SAMPLES samples per call to snd_render(), but
	 * that doesn't seem to correspond to the frame or packet size.
	 * So maybe I need to do another AudioUnitSetProperty() for the
	 * buffer size? */
	sfmt.mBytesPerPacket = 4;
	sfmt.mFramesPerPacket = 1;
	sfmt.mBytesPerFrame = 4;
	sfmt.mChannelsPerFrame = 2;
	sfmt.mBitsPerChannel = 16;

	err = AudioUnitSetProperty (snd_out_unit,
		kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input,
		0, &sfmt, sizeof sfmt);
	if (err) {
		fprintf(stderr, "AudioUnitSetProperty-SF=%4.4s, %ld\n", (char*)&err, err);
		return -1;
	}

	/* initialize unit */
	err = AudioUnitInitialize(snd_out_unit);
	if (err) {
		fprintf(stderr, "AudioUnitInitialize-SF=%4.4s, %d\n", (char*)&err, (int)err);
		return -1;
	}

#if 0
	/* diagnostic the output rate? */
	Float64 outSampleRate;
	UInt32 size = sizeof(Float64);
	err = AudioUnitGetProperty (snd_out_unit,
		kAudioUnitProperty_SampleRate, kAudioUnitScope_Output,
		0, &outSampleRate, &size);
#endif

	AudioOutputUnitStart(snd_out_unit);

	snd_fd = 0;		/* just so we know it's inited */

	return 0;
}

static signed short *
snd_dobuf(void)
{
	signed short *sndbuf = snd_compute();
#if 0
	/* XXX */
	if (mic_level) {
		snd_read_mic(sndbuf);
	} else if (mic_fd != -1) {
		close(mic_fd);
		mic_fd = -1;
	}
#endif

	if (snd_file_fd != -1) {
		write(snd_file_fd, sndbuf, FRAG_SAMPLES*4);
	}

	return sndbuf;
}

OSStatus snd_render(void *inRefCon,
	AudioUnitRenderActionFlags *ioActionFlags,
	const AudioTimeStamp *inTimeStamp,
	UInt32 inBusNumber,
	UInt32 inNumberFrames,
	AudioBufferList *ioData)
{
	static signed short *sndbuf = NULL;
	static int sndbuf_ofs = 0;

	if (!snd_enabled || (snd_fd == -1)) {
		return noErr;
	}

	while (inNumberFrames) {
		int tocopy;

		if ((sndbuf_ofs >= (FRAG_SAMPLES*4)) ||
		    (!sndbuf)) {
			sndbuf = snd_dobuf();
			sndbuf_ofs = 0;
		}

			/* output to the audio unit buffer... */
		tocopy = MIN(inNumberFrames*4, (FRAG_SAMPLES*4) - sndbuf_ofs);

		memcpy(ioData->mBuffers[0].mData, ((char *)sndbuf)+sndbuf_ofs, tocopy);
		inNumberFrames -= tocopy/4;
		sndbuf_ofs += tocopy;
	}

	return noErr;
}


void
snd_out(void)
{
	if (snd_fd != -1) {
		/* do nothing - a magic thread handles sound output, not
		 * the select loop which calls snd_out() */
	} else {
		/* must be writing to snd_file_fd or something */
		snd_dobuf();
	}
}

void
snd_off(void)
{
	AudioOutputUnitStop(snd_out_unit);
	AudioUnitUninitialize(snd_out_unit);
	CloseComponent(snd_out_unit);

	snd_fd = -1;
	/* XXX: close the mic stuff? */
}

void
snd_changedev(char *dev)
{
	fprintf(stderr, "Currently only supports \"default output unit\" in CoreAudio.\n");
}

#else

/* dummy for timing or whatever: */


int
snd_init(void)
{
	snd_fd = 1;
	sounds = NULL;

	return 0;
}

void
snd_out(void)
{
	(void)snd_compute();
	if (sounds == NULL) {
		do_cleanup();
		exit(0);
	}
}

void
snd_cleanup(void)
{
	snd_shutup();
}

void
snd_changedev(char *dev)
{
}


#endif

int
snd_init(void)
{
	sounds = NULL;

	return snd_on();
}

void
snd_cleanup(void)
{
	snd_shutup();

	snd_off();

	if (snd_file_fd != -1) close(snd_file_fd); snd_file_fd = -1;
}



/* silence all currently active sounds, for example because the environment
 * changed */
void
snd_shutup(void)
{
	struct sound_rec *s,*s_next;

	for (s = sounds; s; s = s_next) {
		s_next = s->next;
		free_tree(s->tr);
		free(s);
	}
	sounds = NULL;
}

/* make the note with the specified id register a noteup event, if it
 * is currently playing etc */
void
snd_noteup(int id)
{
	struct sound_rec *s;
	for (s = sounds; s; s = s->next) {
		if (s->id == id) {
			s->note_up = s->ofs;
			break;
		}
	}
}

static int frag_pos = -1;	/* when computing inside a fragment, this is the
                         	 * offset (or -1 for outside of frag) */

void
snd_add(struct tree_rec *tr, value_t *tmps, intval_t *itmps)
{
	static int sound_ids = 1;
	struct sound_rec *s;

	s = (struct sound_rec *)safe_malloc(sizeof *s);
	s->next = sounds;
	s->tr = tr;
	s->ofs = 0;
	s->note_up = 44100*60;	/* make sure no note lasts forever */
	s->id = sound_ids++;
	s->tmps = tmps;
	s->itmps = itmps;
	if (frag_pos != -1) {
		/* computing a frame, go ahead and fill this one in */
		curr_sound = s;
		if (do_fn_call(s, snd_fbuf+(frag_pos*2), FRAG_SAMPLES-frag_pos)) {
			/* done already?  wow! */
			free(curr_sound);
			return;
		}
	}

	sounds = s;
}


static signed short *
snd_compute(void)
{
	static signed short *sndbuf = NULL;
	struct sound_rec **sp;
	int i;

	if (!sndbuf) {
		sndbuf = safe_malloc(FRAG_SAMPLES*2*sizeof *sndbuf);
		snd_fbuf = safe_malloc(FRAG_SAMPLES*2*sizeof *snd_fbuf);
	}

	memset(snd_fbuf, 0, FRAG_SAMPLES*2*sizeof *snd_fbuf);

	for (sp = &sounds; *sp; ) {
		curr_sound = *sp;
		if (do_fn_call(curr_sound, snd_fbuf, FRAG_SAMPLES)) {
				/* delete this sound */
			*sp = curr_sound->next;
			free(curr_sound);
			curr_sound = NULL;
		} else {
			sp = &(curr_sound->next);
		}
	}
	for (i = 0; i < FRAG_SAMPLES; i++) {
		int l,r;

		frag_pos = i;
		tracker_timestep();
		midi_play_step();
		pattern_step();
		gyro_step();
		frag_pos = -1;

		l = snd_fbuf[i*2+0];
		r = snd_fbuf[i*2+1];

		if (l < -32767) l = -32767;
		if (l > 32767) l = 32767;
		if (r < -32767) r = -32767;
		if (r > 32767) r = 32767;

		if (abs(l) > snd_max) snd_max = abs(l);
		if (abs(r) > snd_max) snd_max = abs(r);

			/* now save it in the actual buffer */
		sndbuf[i*2+0] = l;
		sndbuf[i*2+1] = r;
	}

	return sndbuf;
}
