/* Copyright 2010 Greg Alexander.  See the file COPYING for terms */
/* $Id: snd.h,v 1.26 2010-02-08 03:17:28 greg Exp $ */
#ifndef SND_H
#define SND_H


#include "misc.h"
#include "lang.h"

#if DEFINE_GLOBALS
#define __EXTERN__
#define EQUALS(x) = (x)
#else
#define __EXTERN__ extern
#define EQUALS(x)
#endif

#if defined(ALSA_SOUND)
# if defined(MAEMO)
#  define ALSA_FILE "default"
# else
#  define ALSA_FILE "plughw:0,0"
# endif
#elif defined(OSS_SOUND)
# define SND_FILE "/dev/dsp"
/* FRAG_LEN may need performance tuning:
 *   12 => 1024 samples => 23ms    (conservative)
 *   10 => 256 samples => 6ms
 *    7 => 32 samples => .7ms
 */
# define FRAG_LEN 10		/* 2^7 => 128 bytes per fragment */
# define FRAG_SAMPLES ((1<<FRAG_LEN)/4)
# define MIC_FILE "/dev/dsp0"
#elif defined (COREAUDIO_SOUND)
# define FRAG_SAMPLES 1024
#else /* dummy */
# define FRAG_SAMPLES 1024
#endif

	/* !!!         do not change sound_rec         !!! */
	/* !!! unless you check compile.c's definition !!! */
struct sound_rec {
        struct sound_rec *next;
	struct tree_rec *tr;
	unsigned long ofs;
	unsigned long note_up;
	int id;			/* for more general note up */
	value_t *tmps;
	intval_t *itmps;
};

__EXTERN__ struct sound_rec *sounds EQUALS(NULL);
__EXTERN__ struct sound_rec *curr_sound EQUALS(NULL);
__EXTERN__ int snd_fd EQUALS(-1);	/* the /dev for output */
__EXTERN__ int snd_file_fd EQUALS(-1);	/* an auxiliary output */
__EXTERN__ int mic_level EQUALS(0);	/* between 0 and 32767 */
__EXTERN__ unsigned snd_max EQUALS(0);	/* between 0 and 32767 */
__EXTERN__ int snd_enabled EQUALS(1);	/* master on/off switch */

/* snd.c: */
int snd_init(void);
int snd_on(void);
void snd_off(void);
void snd_add(struct tree_rec *tr, value_t *tmps, intval_t *itmps);
void snd_out(void);
void snd_cleanup(void);
void snd_shutup(void);
void snd_noteup(int id);
void snd_changedev(char *dev);


#undef __EXTERN__
#undef EQUALS

#endif /* SND_H */
